package com.jiuqi.grid;

import com.jiuqi.util.Base64;
import com.jiuqi.util.MemStream;
import com.jiuqi.util.Stream;
import com.jiuqi.util.StreamException;
import com.jiuqi.util.StringUtils;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Stack;

public final class GridData implements Serializable, Cloneable {
   private IntList cellColors;
   private IntList fontColors;
   private IntList edgeColors;
   private GridFontNames fontNames;
   private IntList fontSizes;
   private IntList colWidths;
   private IntList rowHeights;
   private BitList colVisible;
   private BitList rowVisible;
   private BitList hideDataRow;
   private BitList hideDataCol;
   private BitList colAutoSize;
   private BitList rowAutoSize;
   private Matrix values;
   private Matrix cells;
   private GridFieldList rects;
   private GridCell cell;
   private PropList propList;
   private Matrix objects;
   private int scrollTopCol;
   private int scrollTopRow;
   private int scrollBottomCol;
   private int scrollBottomRow;
   private int curCol;
   private int curRow;
   private boolean useFontSize;
   private LabelItems labels;
   private Matrix cellCSS;
   private String gridDataClass;
   private String referenceBase;
   private Matrix cellScript;
   private int horzOffset;
   private int vertOffset;
   private boolean horzCenter;
   private boolean vertCenter;
   private ExtendedDatas extDatas;
   private int options;
   private Set modifiedPropIndexRecycle;
   private boolean modifying;
   private int colCountPerPage;
   private int rowCountPerPage;
   private boolean rowPagingFirst;
   public static final int GRID_AUTOMERGE = 1;
   public static final int GRID_AUTOSTYLE = 2;
   public static final int GRID_DYNAMICSTORE = 4;
   public static final int GRID_FULLSTYLE = 8;

   public GridData() {
      this.scrollTopCol = -1;
      this.scrollTopRow = -1;
      this.scrollBottomCol = -1;
      this.scrollBottomRow = -1;
      this.curCol = -1;
      this.curRow = -1;
      this.labels = new LabelItems();
      this.cellCSS = null;
      this.cellScript = null;
      this.modifying = false;
      this.colCountPerPage = 0;
      this.rowCountPerPage = 0;
      this.rowPagingFirst = true;
      this.cellColors = GridConsts.create64ColorsPallette();
      this.fontColors = GridConsts.create64ColorsPallette();
      this.edgeColors = GridConsts.create16ColorsPallette();
      this.fontNames = new GridFontNames(64);
      this.fontNames.addFont("宋体");
      this.fontSizes = new IntList();
      this.fontSizes.add(9);
      this.colVisible = new BitList(true);
      this.rowVisible = new BitList(true);
      this.colAutoSize = new BitList(false);
      this.rowAutoSize = new BitList(false);
      this.hideDataRow = new BitList(false);
      this.hideDataCol = new BitList(false);
      this.colWidths = new IntList();
      this.rowHeights = new IntList();
      this.values = new Matrix();
      this.cells = new Matrix();
      this.objects = new Matrix();
      this.rects = new GridFieldList(this);
      this.cell = new GridCell();
      this.propList = new PropList(20);
      this.propList.add(GridConsts.DEF_CELL_PROP);
      this.extDatas = new ExtendedDatas();
      this.setColCount(5);
      this.setRowCount(5);

      int i;
      GridCell cell;
      for(i = 0; i < 5; ++i) {
         cell = this.getCell(i, 0);
         this.setCell(i, 0, cell);
      }

      for(i = 0; i < 5; ++i) {
         cell = this.getCell(0, i);
         this.setCell(0, i, cell);
      }

      this.options = 7;
      this.setOptions(this.options);
   }

   public GridData(int colCount, int rowCount) {
      this();
      this.setColCount(colCount);
      this.setRowCount(rowCount);
   }

   public GridData(int options) {
      this();
      this.setOptions(options);
   }

   public GridData(int colCount, int rowCount, int options) {
      this(colCount, rowCount);
      this.setOptions(options);
   }

   public void beginModify() {
      if (this.isModifying()) {
         throw new IllegalStateException("无法对已经处于修改状态的GridData执行beginModify方法！");
      } else {
         if (this.modifiedPropIndexRecycle == null) {
            this.modifiedPropIndexRecycle = new HashSet();
         }

         this.modifying = true;
      }
   }

   public void secureBeginModify() {
      synchronized(this) {
         this.beginModify();
      }
   }

   public boolean isModifying() {
      return this.modifying;
   }

   public void endModify() {
      this.rebuildPropIndexes(this.modifiedPropIndexRecycle);
      this.modifiedPropIndexRecycle.clear();
      this.modifying = false;
   }

   public void secureEndModify() {
      synchronized(this) {
         this.endModify();
      }
   }

   private void createCellCssMatrix() {
      if (this.cellCSS == null) {
         this.cellCSS = new Matrix();
         this.cellCSS.setSize(this.getColCount(), this.getRowCount());
      }

   }

   private void createCellScriptMatrix() {
      if (this.cellScript == null) {
         this.cellScript = new Matrix();
         this.cellScript.setSize(this.getColCount(), this.getRowCount());
      }

   }

   private boolean checkCol(int col, boolean isInsert) {
      if (isInsert) {
         return col > 0 && col <= this.getColCount();
      } else {
         return col >= 0 && col < this.getColCount();
      }
   }

   private boolean checkRow(int row, boolean isInsert) {
      if (isInsert) {
         return row > 0 && row <= this.getRowCount();
      } else {
         return row >= 0 && row < this.getRowCount();
      }
   }

   private void rebuildPropIndex(int col, int row, int oldIndex) {
      if (oldIndex >= 0) {
         if (this.modifying) {
            this.modifiedPropIndexRecycle.add(new Integer(oldIndex));
         } else {
            if (!this.cells.hasElement(new Integer(oldIndex))) {
               for(int x = 0; x < this.cells.getXCount(); ++x) {
                  for(int y = 0; y < this.cells.getYCount(); ++y) {
                     Integer index = (Integer)this.cells.getElement(x, y);
                     if (index != null && index > oldIndex) {
                        index = new Integer(index - 1);
                        this.cells.setElement(x, y, index);
                     }
                  }
               }

               this.propList.remove(oldIndex);
            }

         }
      }
   }

   private void rebuildPropIndexes(Set bufferedIndex) {
      if (bufferedIndex.size() != 0 && this.propList.count() != 0) {
         int[] propCase = new int[this.propList.count()];
         Iterator indexIt = bufferedIndex.iterator();

         while(indexIt.hasNext()) {
            Object index = indexIt.next();
            if (!this.cells.hasElement(index)) {
               propCase[(Integer)index] = -1;
            }
         }

         int x;
         for(x = propCase.length - 1; x >= 0; --x) {
            if (propCase[x] != 0) {
               this.propList.remove(x);
            }
         }

         for(x = 1; x < propCase.length; ++x) {
            propCase[x] += propCase[x - 1];
         }

         if (propCase[propCase.length - 1] != 0) {
            for(x = 0; x < this.cells.getXCount(); ++x) {
               for(int y = 0; y < this.cells.getYCount(); ++y) {
                  Integer index = (Integer)this.cells.getElement(x, y);
                  if (index != null && propCase[index] != 0) {
                     index = new Integer(index + propCase[index]);
                     this.cells.setElement(x, y, index);
                  }
               }
            }

         }
      }
   }

   public String getDataClass() {
      return this.gridDataClass;
   }

   public void setDataClass(String value) {
      this.gridDataClass = value;
   }

   public String getReferenceBase() {
      return this.referenceBase;
   }

   public void setReferenceBase(String value) {
      this.referenceBase = value;
   }

   public String getCellData(int col, int row) {
      return (String)this.values.getElement(col, row);
   }

   public String getCellData(String pos) {
      CellPostion p = new CellPostion(pos);
      return this.getCellData(p.getCol(), p.getRow());
   }

   public void setCellData(int col, int row, String value) {
      this.values.setElement(col, row, value);
   }

   public void setCellData(String pos, String value) {
      CellPostion p = new CellPostion(pos);
      this.setCellData(p.getCol(), p.getRow(), value);
   }

   public String getCellClass(int col, int row) {
      return this.cellCSS == null ? null : (String)this.cellCSS.getElement(col, row);
   }

   public void setCellClass(int col, int row, String value) {
      if (this.cellCSS == null) {
         if (value == null || value.length() == 0) {
            return;
         }

         this.createCellCssMatrix();
      }

      this.cellCSS.setElement(col, row, value);
   }

   public boolean getUseFontSize() {
      return this.useFontSize;
   }

   public void setUserFontSize(boolean value) {
      this.useFontSize = value;
   }

   public String getCellScript(int col, int row) {
      return this.cellScript == null ? null : (String)this.cellScript.getElement(col, row);
   }

   public void setCellScript(int col, int row, String value) {
      if (this.cellScript == null) {
         if (value == null || value.length() == 0) {
            return;
         }

         this.createCellScriptMatrix();
      }

      this.cellScript.setElement(col, row, value);
   }

   public int getCellStyleIndex(int col, int row) {
      Integer index = (Integer)this.cells.getElement(col, row);
      return index == null ? 0 : index;
   }

   public GridCell[] getCellSytles() {
      if (this.propList.count() <= 0) {
         return new GridCell[0];
      } else {
         GridCell[] ret = new GridCell[this.propList.count()];

         for(int i = 0; i < ret.length; ++i) {
            ret[i] = new GridCell();
            ret[i].init(this, 0, 0, this.propList.get(i));
         }

         return ret;
      }
   }

   public GridCell[] getCellStyles(int left, int top, int right, int bottom, boolean fullMode) {
      if (this.propList.count() == 0) {
         return new GridCell[0];
      } else {
         GridCell[] ret = new GridCell[this.propList.count()];

         for(int col = left; col <= right; ++col) {
            for(int row = top; row <= bottom; ++row) {
               if (StringUtils.isEmpty(this.getCellClass(col, row))) {
                  int index = this.getCellStyleIndex(col, row);
                  if (ret[index] == null) {
                     ret[index] = new GridCell();
                     if (fullMode) {
                        ret[index].init(this, col, row);
                     } else {
                        ret[index].init(this, 0, 0, this.getCellPropData(col, row));
                     }
                  }
               }
            }
         }

         return ret;
      }
   }

   public byte[] getCellPropData(int col, int row) {
      Integer index = (Integer)this.cells.getElement(col, row);
      if (index != null) {
         return this.propList.get(index);
      } else {
         return col != 0 && row != 0 ? GridConsts.DEF_CELL_PROP : GridConsts.DEF_HEADCELL_PROP;
      }
   }

   public GridCell getCell(int col, int row) {
      this.cell.init(this, col, row);
      return this.cell;
   }

   public GridCell getCell(String pos) {
      CellPostion p = new CellPostion(pos);
      return this.getCell(p.getCol(), p.getRow());
   }

   public GridCell getCellEx(int col, int row) {
      GridCell result = new GridCell();
      result.init(this, col, row);
      return result;
   }

   public GridCell getCellEx(String pos) {
      CellPostion p = new CellPostion(pos);
      return this.getCellEx(p.getCol(), p.getRow());
   }

   public GridCellProperty getCellForChange(int col, int row) {
      if (this.checkCol(col, false) && this.checkRow(row, false)) {
         GridCellProperty result = new GridCellProperty();
         result.init(this, col, row);
         return result;
      } else {
         return null;
      }
   }

   public GridCellProperty getCellForChange(String pos) {
      CellPostion p = new CellPostion(pos);
      return this.getCellForChange(p.getCol(), p.getRow());
   }

   public void setCell(int col, int row, GridCell value) {
      int index = this.propList.indexOf(value.getPropData());
      if (index < 0) {
         index = this.propList.add(value.getPropData());
      }

      boolean dynamicStore = (this.options & 4) != 0;
      Integer oldIndex = dynamicStore ? (Integer)this.cells.getElement(col, row) : null;
      this.cells.setElement(col, row, new Integer(index));
      if (dynamicStore && oldIndex != null && oldIndex != index) {
         this.rebuildPropIndex(col, row, oldIndex);
      }

   }

   public void setCell(GridCell value) {
      this.setCell(value.getColNum(), value.getRowNum(), value);
   }

   public void setCell(int left, int top, int right, int bottom, GridCell value) {
      this.setCell(new CellField(left, top, right, bottom), value);
   }

   public void setCell(CellField field, GridCell value) {
      int index = this.propList.indexOf(value.getPropData());
      if (index == -1) {
         index = this.propList.add(value.getPropData());
      }

      Integer p = new Integer(index);

      for(int col = field.left; col <= field.right; ++col) {
         for(int row = field.top; row <= field.bottom; ++row) {
            CellField f = this.rects.getMergeRect(col, row);
            if (f == null || f.left == col && f.top == row) {
               this.cells.setElement(col, row, p);
            }
         }
      }

   }

   public void setCell(CellField[] fields, GridCell value) {
      if (fields != null && fields.length != 0) {
         int index = this.propList.indexOf(value.getPropData());
         if (index == -1) {
            index = this.propList.add(value.getPropData());
         }

         Integer p = new Integer(index);

         for(int i = 0; i < fields.length; ++i) {
            for(int col = fields[i].left; col <= fields[i].right; ++col) {
               for(int row = fields[i].top; row <= fields[i].bottom; ++row) {
                  CellField f = this.rects.getMergeRect(col, row);
                  if (f == null || f.left == col && f.top == row) {
                     this.cells.setElement(col, row, p);
                  }
               }
            }
         }

      }
   }

   public Object getObj(int col, int row) {
      return this.objects.getElement(col, row);
   }

   public void setObj(int col, int row, Object obj) {
      this.objects.setElement(col, row, obj);
   }

   public boolean getColAutoSize(int col) {
      return this.colAutoSize.get(col);
   }

   public void setColAutoSize(int col, boolean value) {
      this.colAutoSize.set(col, value);
   }

   public boolean getColVisible(int col) {
      return this.colVisible.count() > 0 ? this.colVisible.get(col) : true;
   }

   public void setColVisible(int col, boolean value) {
      if (!value || col < this.colVisible.count()) {
         this.colVisible.set(col, value);
      }

   }

   public int getColCount() {
      return this.colWidths.count();
   }

   public int getColWidths(int col) {
      return this.colWidths.get(col);
   }

   public void setColWidths(int col, int value) {
      this.colWidths.set(col, value);
   }

   public boolean getRowAutoSize(int row) {
      return this.rowAutoSize.get(row);
   }

   public void setRowAutoSize(int row, boolean value) {
      this.rowAutoSize.set(row, value);
   }

   public boolean getRowVisible(int row) {
      return this.rowVisible.count() > 0 ? this.rowVisible.get(row) : true;
   }

   public void setRowVisible(int row, boolean value) {
      if (!value || row < this.getRowCount()) {
         this.rowVisible.set(row, value);
      }

   }

   /** @deprecated */
   public boolean getHideDataRow(int row) {
      return this.hideDataRow.get(row);
   }

   /** @deprecated */
   public void setHideDataRow(int row, boolean value) {
      if (!value || row < this.getRowCount()) {
         this.hideDataRow.set(row, value);
      }

   }

   /** @deprecated */
   public boolean getHideDataCol(int col) {
      return this.hideDataCol.get(col);
   }

   /** @deprecated */
   public int[][] getHideDataCols() {
      int[][] result = (int[][])null;
      int count = 0;
      boolean state = false;

      int k;
      for(k = 1; k < this.getColCount(); ++k) {
         if (!state && this.getHideDataCol(k)) {
            ++count;
            state = true;
         }

         if (state && !this.getHideDataCol(k)) {
            state = false;
         }
      }

      if (count > 0) {
         result = new int[2][count];
         state = false;
         k = 0;

         for(int i = 1; i < this.getColCount(); ++i) {
            if (!state && this.getHideDataCol(i)) {
               result[0][k] = i;
               state = true;
            }

            if (state && !this.getHideDataCol(i)) {
               state = false;
               result[1][k] = i - 1;
               ++k;
            }
         }

         if (state) {
            result[1][count - 1] = this.getColCount() - 1;
         }
      }

      return result;
   }

   /** @deprecated */
   public int[][] getHideDataRows() {
      int[][] result = (int[][])null;
      int count = 0;
      boolean state = false;

      int k;
      for(k = 1; k < this.getRowCount(); ++k) {
         if (!state && this.getHideDataRow(k)) {
            ++count;
            state = true;
         }

         if (state && !this.getHideDataRow(k)) {
            state = false;
         }
      }

      if (count > 0) {
         result = new int[2][count];
         state = false;
         k = 0;

         for(int i = 1; i < this.getRowCount(); ++i) {
            if (!state && this.getHideDataRow(i)) {
               result[0][k] = i;
               state = true;
            }

            if (state && !this.getHideDataRow(i)) {
               state = false;
               result[1][k] = i - 1;
               ++k;
            }
         }

         if (state) {
            result[1][count - 1] = this.getRowCount() - 1;
         }
      }

      return result;
   }

   /** @deprecated */
   public void setHideDataCol(int col, boolean value) {
      if (!value || col < this.getColCount()) {
         this.hideDataCol.set(col, value);
      }

   }

   public int getRowCount() {
      return this.rowHeights.count();
   }

   public void setColCount(int value) {
      int colCount = this.getColCount();
      int insCount;
      if (value > 0 && value < colCount) {
         insCount = colCount - value;
         this.colVisible.delete(value, insCount);
         this.colAutoSize.delete(value, insCount);
         this.colWidths.delete(value, insCount);
         this.values.xDelete(value, insCount);
         this.cells.xDelete(value, insCount);
         this.objects.xDelete(value, insCount);
         if (this.cellCSS != null) {
            this.cellCSS.xDelete(value, insCount);
         }

         if (this.cellScript != null) {
            this.cellScript.xDelete(value, insCount);
         }

         this.rects.delCols(value, insCount);
      } else if (value > colCount) {
         insCount = value - colCount;
         this.colVisible.insert(colCount, insCount);
         this.colAutoSize.insert(colCount, insCount);
         this.colWidths.insert(colCount, insCount, colCount == 0 ? 100 : this.colWidths.get(colCount - 1));
         if (colCount == 0) {
            this.colWidths.set(0, 50);
         }

         this.values.xInsert(colCount, insCount);
         this.cells.xInsert(colCount, insCount);
         this.objects.xInsert(colCount, insCount);
         if (this.cellCSS != null) {
            this.cellCSS.xInsert(colCount, insCount);
         }

         if (this.cellScript != null) {
            this.cellScript.xInsert(colCount, insCount);
         }

         this.rects.insertCols(colCount, insCount);
         if (this.getRowCount() > 0) {
            for(int i = colCount; i < colCount + insCount; ++i) {
               GridCell cell = this.getCell(i, 0);
               this.setCell(i, 0, cell);
            }
         }
      }

      colCount = this.getColCount();
      if (this.scrollTopCol >= colCount) {
         this.setScrollTopCol(colCount - 1);
      }

      if (this.scrollBottomCol >= colCount) {
         this.setScrollBottomCol(colCount - 1);
      }

   }

   public void setRowCount(int value) {
      int rowCount = this.getRowCount();
      int insCount;
      if (value < rowCount) {
         insCount = rowCount - value;
         this.rowVisible.delete(value, insCount);
         this.rowAutoSize.delete(value, insCount);
         this.rowHeights.delete(value, insCount);
         this.values.yDelete(value, insCount);
         this.cells.yDelete(value, insCount);
         this.objects.yDelete(value, insCount);
         if (this.cellCSS != null) {
            this.cellCSS.yDelete(value, insCount);
         }

         if (this.cellScript != null) {
            this.cellScript.yDelete(value, insCount);
         }

         this.rects.delRows(value, insCount);
      } else if (value > this.getRowCount()) {
         insCount = value - rowCount;
         this.rowVisible.insert(rowCount, insCount);
         this.rowAutoSize.insert(rowCount, insCount);
         this.rowHeights.insert(rowCount, insCount, rowCount == 0 ? 20 : this.rowHeights.get(rowCount - 1));
         this.values.yInsert(rowCount, insCount);
         this.objects.yInsert(rowCount, insCount);
         this.cells.yInsert(rowCount, insCount);
         if (this.cellCSS != null) {
            this.cellCSS.yInsert(rowCount, insCount);
         }

         if (this.cellScript != null) {
            this.cellScript.yInsert(rowCount, insCount);
         }

         this.rects.insertRows(rowCount, insCount);
         if (this.getColCount() > 0) {
            for(int i = rowCount; i < rowCount + insCount; ++i) {
               GridCell cell = this.getCell(0, i);
               this.setCell(0, i, cell);
            }
         }
      }

   }

   public void insertCol(int index, int count) {
      this.insertCol(index, count, (this.options & 2) == 0, (this.options & 8) != 0);
   }

   private void insertCol(int index, int count, boolean fastMode, boolean fullMode) {
      if (this.checkCol(index, true)) {
         this.colVisible.insert(index, count);
         if (index > 0) {
            this.colAutoSize.insert(index, count, this.colAutoSize.get(index - 1));
         } else {
            this.colAutoSize.insert(index, count);
         }

         this.colWidths.insert(index, count, this.getColCount() != 0 && index != 1 ? this.colWidths.get(index - 1) : 100);
         this.values.xInsert(index, count);
         this.cells.xInsert(index, count);
         this.objects.xInsert(index, count);
         if (this.cellCSS != null) {
            this.cellCSS.xInsert(index, count);
         }

         if (this.cellScript != null) {
            this.cellScript.xInsert(index, count);
         }

         this.rects.insertCols(index, count);
         if (this.getRowCount() > 0 && !fastMode) {
            for(int row = 1; row < this.getRowCount(); ++row) {
               GridCell cell;
               if (index > 1) {
                  cell = this.getCell(index - 1, row);
               } else {
                  cell = this.cell;
                  cell.init(this, -1, -1, GridConsts.DEF_CELL_PROP);
               }

               for(int col = index; col < index + count; ++col) {
                  this.setCell(col, row, cell);
                  if (fullMode) {
                     this.setCellClass(col, row, cell.getCssClass());
                     this.setCellScript(col, row, cell.getScript());
                  }
               }
            }
         }

         if (this.scrollTopCol >= index) {
            this.scrollTopCol += count;
         }

         if (this.scrollBottomCol + 1 >= index) {
            this.scrollBottomCol += count;
         }

      }
   }

   public void deleteCol(int index, int count) {
      if (this.checkCol(index, false)) {
         if (index + count > this.getColCount()) {
            count = this.getColCount() - index;
         }

         this.colVisible.delete(index, count);
         this.colAutoSize.delete(index, count);
         this.colWidths.delete(index, count);
         this.values.xDelete(index, count);
         this.cells.xDelete(index, count);
         this.objects.xDelete(index, count);
         if (this.cellCSS != null) {
            this.cellCSS.xDelete(index, count);
         }

         if (this.cellScript != null) {
            this.cellScript.xDelete(index, count);
         }

         this.rects.delCols(index, count);
         if (this.scrollTopCol >= index) {
            this.scrollTopCol = Math.min(Math.max(index, this.scrollTopCol - count), this.getColCount() - 1);
         }

         if (this.scrollBottomCol >= index) {
            this.scrollBottomCol = Math.min(Math.max(index, this.scrollBottomCol - count), this.getColCount() - 1);
         }

      }
   }

   public void insertRow(int index, int count) {
      this.insertRow(index, count, (this.options & 2) == 0, (this.options & 8) != 0);
   }

   private void insertRow(int index, int count, boolean fastMode, boolean fullMode) {
      if (this.checkRow(index, true)) {
         this.rowVisible.insert(index, count);
         if (index > 0) {
            this.rowAutoSize.insert(index, count, this.rowAutoSize.get(index - 1));
         } else {
            this.rowAutoSize.insert(index, count);
         }

         this.rowHeights.insert(index, count, this.getRowCount() != 0 && index != 1 ? this.rowHeights.get(index - 1) : 20);
         this.values.yInsert(index, count);
         this.cells.yInsert(index, count);
         this.objects.yInsert(index, count);
         if (this.cellCSS != null) {
            this.cellCSS.yInsert(index, count);
         }

         if (this.cellScript != null) {
            this.cellScript.yInsert(index, count);
         }

         this.rects.insertRows(index, count);
         if (this.getColCount() > 0 && !fastMode) {
            for(int col = 1; col < this.getColCount(); ++col) {
               GridCell cell;
               if (index > 1) {
                  cell = this.getCell(col, index - 1);
               } else {
                  cell = this.cell;
                  cell.init(this, -1, -1, GridConsts.DEF_CELL_PROP);
               }

               for(int row = index; row < index + count; ++row) {
                  this.setCell(col, row, cell);
                  if (fullMode) {
                     this.setCellClass(col, row, cell.getCssClass());
                     this.setCellScript(col, row, cell.getScript());
                  }
               }
            }
         }

         if (this.scrollTopRow >= index) {
            this.scrollTopRow += count;
         }

         if (this.scrollBottomRow + 1 >= index) {
            this.scrollBottomRow += count;
         }

      }
   }

   public boolean duplicateCols(int fromIndex, int toIndex, int duplicateCount, int insertIndex) {
      if (!this.checkCol(fromIndex, false)) {
         return false;
      } else if (!this.checkCol(toIndex, false)) {
         return false;
      } else if (!this.checkCol(insertIndex, true)) {
         return false;
      } else if (toIndex < fromIndex || duplicateCount <= 0 || insertIndex > fromIndex && insertIndex <= toIndex) {
         return false;
      } else {
         int insertCount = (toIndex - fromIndex + 1) * duplicateCount;
         this.insertCol(insertIndex, insertCount, true, true);
         if (insertIndex <= fromIndex) {
            fromIndex += insertCount;
            toIndex += insertCount;
         }

         List merges = fromIndex < toIndex ? this.rects.findFieldsInCols(fromIndex, toIndex) : null;

         for(int i = 0; i < duplicateCount; ++i) {
            int offset = insertIndex + (toIndex - fromIndex + 1) * i - fromIndex;

            int j;
            for(j = fromIndex; j <= toIndex; ++j) {
               this.colWidths.set(offset + j, this.colWidths.get(j));
               this.values.xCopy(j, offset + j);
               if (this.cellCSS != null) {
                  this.cellCSS.xCopy(j, offset + j);
               }

               if (this.cellScript != null) {
                  this.cellScript.xCopy(j, offset + j);
               }

               this.cells.xCopy(j, offset + j);
               if (!this.colVisible.get(j)) {
                  this.colVisible.set(j + offset, false);
               }
            }

            if (merges != null && !merges.isEmpty()) {
               for(j = 0; j < merges.size(); ++j) {
                  CellField cf = (CellField)merges.get(j);
                  this.mergeCells(cf.left + offset, cf.top, cf.right + offset, cf.bottom);
               }
            }
         }

         return true;
      }
   }

   public boolean duplicateRows(int fromIndex, int toIndex, int duplicateCount, int insertIndex) {
      if (!this.checkRow(fromIndex, false)) {
         return false;
      } else if (!this.checkRow(toIndex, false)) {
         return false;
      } else if (!this.checkRow(insertIndex, true)) {
         return false;
      } else if (toIndex < fromIndex || duplicateCount <= 0 || insertIndex > fromIndex && insertIndex <= toIndex) {
         return false;
      } else {
         int insertCount = (toIndex - fromIndex + 1) * duplicateCount;
         this.insertRow(insertIndex, insertCount, true, true);
         if (insertIndex <= fromIndex) {
            fromIndex += insertCount;
            toIndex += insertCount;
         }

         List merges = fromIndex < toIndex ? this.rects.findFieldsInRows(fromIndex, toIndex) : null;

         for(int i = 0; i < duplicateCount; ++i) {
            int offset = insertIndex + (toIndex - fromIndex + 1) * i - fromIndex;

            int j;
            for(j = fromIndex; j <= toIndex; ++j) {
               this.rowHeights.set(offset + j, this.rowHeights.get(j));
               this.values.yCopy(j, offset + j);
               if (this.cellCSS != null) {
                  this.cellCSS.yCopy(j, offset + j);
               }

               if (this.cellScript != null) {
                  this.cellScript.yCopy(j, offset + j);
               }

               this.cells.yCopy(j, offset + j);
               if (!this.rowVisible.get(j)) {
                  this.rowVisible.set(j + offset, false);
               }
            }

            if (merges != null && !merges.isEmpty()) {
               for(j = 0; j < merges.size(); ++j) {
                  CellField cf = (CellField)merges.get(j);
                  this.mergeCells(cf.left, cf.top + offset, cf.right, cf.bottom + offset);
               }
            }
         }

         return true;
      }
   }

   public void deleteRow(int index, int count) {
      if (this.checkRow(index, false)) {
         if (index + count > this.getRowCount()) {
            count = this.getRowCount() - index;
         }

         this.rowVisible.delete(index, count);
         this.rowAutoSize.delete(index, count);
         this.rowHeights.delete(index, count);
         this.values.yDelete(index, count);
         this.cells.yDelete(index, count);
         if (this.cellCSS != null) {
            this.cellCSS.yDelete(index, count);
         }

         if (this.cellScript != null) {
            this.cellScript.yDelete(index, count);
         }

         this.objects.yDelete(index, count);
         this.rects.delRows(index, count);
         if (this.scrollTopRow >= index) {
            this.scrollTopRow = Math.min(Math.max(index, this.scrollTopRow - count), this.getRowCount() - 1);
         }

         if (this.scrollBottomRow >= index) {
            this.scrollBottomRow = Math.min(Math.max(index, this.scrollBottomRow - count), this.getRowCount() - 1);
         }

      }
   }

   public int getRowHeights(int row) {
      return this.rowHeights.get(row);
   }

   public void setRowHeights(int row, int value) {
      this.rowHeights.set(row, value);
   }

   public int getScrollTopCol() {
      return this.scrollTopCol;
   }

   public void setScrollTopCol(int value) {
      this.scrollTopCol = value;
   }

   public int getScrollTopRow() {
      return this.scrollTopRow;
   }

   public void setScrollTopRow(int value) {
      this.scrollTopRow = value;
   }

   public int getScrollBottomCol() {
      return this.scrollBottomCol;
   }

   public void setScrollBottomCol(int value) {
      this.scrollBottomCol = value;
   }

   public int getScrollBottomRow() {
      return this.scrollBottomRow;
   }

   public void setScrollBottomRow(int value) {
      this.scrollBottomRow = value;
   }

   /** @deprecated */
   public int lockTopCol() {
      return this.scrollTopCol;
   }

   /** @deprecated */
   public void setLockTopCol(int value) {
      this.scrollTopCol = value;
   }

   /** @deprecated */
   public int lockTopRow() {
      return this.scrollTopRow;
   }

   /** @deprecated */
   public void setLockTopRow(int value) {
      this.scrollTopRow = value;
   }

   /** @deprecated */
   public int lockBottomCol() {
      return this.scrollBottomCol;
   }

   /** @deprecated */
   public void setLockBottomCol(int value) {
      this.scrollBottomCol = value;
   }

   /** @deprecated */
   public int lockBottomRow() {
      return this.scrollBottomRow;
   }

   /** @deprecated */
   public void setLockBottomRow(int value) {
      this.scrollBottomRow = value;
   }

   public int curCol() {
      return this.curCol;
   }

   public void setCurCol(int value) {
      this.curCol = value;
   }

   public int curRow() {
      return this.curRow;
   }

   public void setCurRow(int value) {
      this.curRow = value;
   }

   public GridFieldList merges() {
      return this.rects;
   }

   public IntList cellColors() {
      return this.cellColors;
   }

   public IntList fontColors() {
      return this.fontColors;
   }

   public IntList edgeColors() {
      return this.edgeColors;
   }

   public void setEdgeColor(int index, int color) {
      if (index >= 0 && index < 16) {
         this.edgeColors.set(index, color);
      }

   }

   public GridFontNames fontNames() {
      return this.fontNames;
   }

   public IntList fontSizes() {
      return this.fontSizes;
   }

   public boolean cellMerged(int col, int row) {
      return this.rects.getMergeRect(col, row) != null;
   }

   public CellField expandCell(int col, int row) {
      CellField rt = this.rects.getMergeRect(col, row);
      return rt == null ? new CellField(col, row, col, row) : (CellField)rt.clone();
   }

   public boolean mergeCells(int left, int top, int right, int bottom) {
      CellField cf = new CellField(left, top, right, bottom);
      return this.rects.addMergeRect(cf);
   }

   public LabelItems getLabels() {
      return this.labels;
   }

   private void clear() {
      this.scrollTopCol = -1;
      this.scrollTopRow = -1;
      this.scrollBottomCol = -1;
      this.scrollBottomRow = -1;
      this.curCol = -1;
      this.curRow = -1;
      this.labels.clear();
      this.extDatas.clear();
      this.cellCSS = null;
      this.cellScript = null;
   }

   public void loadFromStream(Stream stream) throws StreamException {
      ValueList valueList = new ValueList();
      ValueList cellCSSList = new ValueList();
      ValueList cellScriptList = new ValueList();
      WordList propIndexs = new WordList();
      KeyValues dataProps = new KeyValues();
      BIFF b = new BIFF();
      b.data().setUseEncode(stream.getUseEncode());
      b.data().setCharset(stream.getCharset());
      long p = stream.getPosition();
      BIFF.readBIFF(1, b, stream);
      String signFlag = b.data().readString(4);
      if (!"BIFF".equals(signFlag)) {
         stream.setPosition(p);
         BIFF.readBIFF(2, b, stream);
         signFlag = b.data().readString(4);
         if (!"BIFF".equals(signFlag)) {
            throw new StreamException("错误的文件格式");
         }
      }

      int ver = b.data().readInt();
      int fileVer;
      if (b.data().getPosition() < b.data().getSize()) {
         fileVer = b.data().readInt();
      } else {
         fileVer = 16777216;
      }

      this.clear();
      this.useFontSize = false;

      while(b.ident != 127) {
         BIFF.readBIFF(ver, b, stream);
         switch (b.ident) {
            case 1:
               propIndexs.loadFromStream(b.data(), b.size);
               break;
            case 2:
               this.colWidths.loadFromStream(b.data(), b.size);
               break;
            case 3:
               this.rowHeights.loadFromStream(b.data(), b.size);
               break;
            case 4:
               valueList.loadFromStream(b.data());
               break;
            case 5:
               this.rects.loadFromStream(b.data(), b.size);
               break;
            case 6:
               this.fontNames.loadFromStream(b.data());
               break;
            case 7:
               this.fontColors.loadFromStream(b.data(), b.size);
               break;
            case 8:
               this.fontSizes.loadFromStream(b.data(), b.size);
               break;
            case 9:
               this.edgeColors.loadFromStream(b.data(), b.size);
               break;
            case 10:
               this.cellColors.loadFromStream(b.data(), b.size);
               break;
            case 11:
               this.propList.loadFromStreamEx(b.data(), 0, 12);
               break;
            case 12:
               this.colVisible.loadFromStream(b.data(), b.size);
               break;
            case 13:
               this.colAutoSize.loadFromStream(b.data(), b.size);
               break;
            case 14:
               this.rowVisible.loadFromStream(b.data(), b.size);
               break;
            case 15:
               this.rowAutoSize.loadFromStream(b.data(), b.size);
               break;
            case 16:
               this.scrollTopCol = b.data().readInt();
               this.scrollTopRow = b.data().readInt();
               this.scrollBottomCol = b.data().readInt();
               this.scrollBottomRow = b.data().readInt();
               break;
            case 17:
               this.curCol = b.data().readInt();
               this.curRow = b.data().readInt();
               break;
            case 18:
               this.useFontSize = true;
               break;
            case 19:
               this.propList.loadFromStreamEx(b.data(), 12, 8);
               break;
            case 20:
               this.labels.loadFromStream(b.data());
               if (fileVer == 16777216) {
                  this.labels.ver10Tover11();
               }
               break;
            case 21:
               this.hideDataCol.loadFromStream(b.data(), b.size);
               break;
            case 22:
               this.hideDataRow.loadFromStream(b.data(), b.size);
               break;
            case 23:
               this.colCountPerPage = b.data().readInt();
               this.rowCountPerPage = b.data().readInt();
               this.rowPagingFirst = b.data().readBool();
               break;
            case 24:
               cellCSSList.loadFromStream(b.data());
               break;
            case 25:
               dataProps.loadFromStream(b.data());
               break;
            case 26:
               cellScriptList.loadFromStream(b.data());
               break;
            case 27:
               this.extDatas.loadFromStream(b.data(), b.size);
         }
      }

      this.gridDataClass = dataProps.getValue("gridclass");
      this.referenceBase = dataProps.getValue("gridrefbase");
      this.horzOffset = dataProps.getAsInt("horzoffset");
      this.vertOffset = dataProps.getAsInt("vertoffset");
      this.horzCenter = dataProps.getAsBoolean("horzcenter");
      this.vertCenter = dataProps.getAsBoolean("vertcenter");
      this.setColCount(this.colWidths.count());
      this.setRowCount(this.rowHeights.count());
      this.cells.resetSize(this.colWidths.count(), this.rowHeights.count());
      this.objects.resetSize(this.colWidths.count(), this.rowHeights.count());
      this.values.resetSize(this.colWidths.count(), this.rowHeights.count());
      int i = 0;

      for(int c = 0; c < this.colWidths.count(); ++c) {
         for(int r = 0; r < this.rowHeights.count(); ++r) {
            this.cells.setElement(c, r, new Integer(propIndexs.get(i)));
            ++i;
         }
      }

      for(i = 0; i < valueList.count(); ++i) {
         this.values.setElement(valueList.col(i), valueList.row(i), valueList.get(i));
      }

      for(i = 0; i < cellCSSList.count(); ++i) {
         this.setCellClass(cellCSSList.col(i), cellCSSList.row(i), cellCSSList.get(i));
      }

      for(i = 0; i < cellScriptList.count(); ++i) {
         this.setCellScript(cellScriptList.col(i), cellScriptList.row(i), cellScriptList.get(i));
      }

   }

   public void saveToStream(Stream stream) throws StreamException {
      ValueList valueList = new ValueList();
      ValueList cellCSSList = new ValueList();
      ValueList cellScriptList = new ValueList();
      WordList propIndexs = new WordList();
      KeyValues dataProps = new KeyValues();
      dataProps.setValue("gridclass", this.gridDataClass);
      dataProps.setValue("gridrefbase", this.referenceBase);
      dataProps.setAsInt("horzoffset", this.horzOffset);
      dataProps.setAsInt("vertoffset", this.vertOffset);
      dataProps.setAsBoolean("horzcenter", this.horzCenter);
      dataProps.setAsBoolean("vertcenter", this.vertCenter);

      for(int c = 0; c < this.colWidths.count(); ++c) {
         for(int r = 0; r < this.rowHeights.count(); ++r) {
            String cell = (String)this.values.getElement(c, r);
            if (cell != null && cell.length() > 0) {
               CellValue cv = new CellValue();
               cv.col = c;
               cv.row = r;
               cv.value = cell;
               valueList.add(cv);
            }

            Integer propIndex = (Integer)this.cells.getElement(c, r);
            if (propIndex == null) {
               int pi = this.propList.indexOf(c != 0 && r != 0 ? GridConsts.DEF_CELL_PROP : GridConsts.DEF_HEADCELL_PROP);
               if (c != 0 && r != 0) {
                  pi = this.propList.add(GridConsts.DEF_CELL_PROP);
               } else {
                  pi = this.propList.add(GridConsts.DEF_HEADCELL_PROP);
               }

               propIndex = new Integer(pi);
            }

            propIndexs.add(propIndex);
            cell = this.getCellClass(c, r);
            CellValue cv;
            if (cell != null && cell.length() > 0) {
               cv = new CellValue();
               cv.col = c;
               cv.row = r;
               cv.value = cell;
               cellCSSList.add(cv);
            }

            cell = this.getCellScript(c, r);
            if (cell != null && cell.length() != 0) {
               cv = new CellValue();
               cv.col = c;
               cv.row = r;
               cv.value = cell;
               cellScriptList.add(cv);
            }
         }
      }

      BIFF b = new BIFF();
      b.ident = 0;
      b.data().writeString("BIFF");
      b.data().writeInt(2);
      b.data().writeInt(16842752);
      BIFF.writeBIFF(1, b, stream);
      b.reset();
      b.ident = 25;
      dataProps.saveToStream(b.data());
      BIFF.writeBIFF(25, b, stream);
      b.reset();
      b.ident = 18;
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 6;
      this.fontNames.saveToStream(b.data());
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 8;
      this.fontSizes.saveToStream(b.data());
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 7;
      this.fontColors.saveToStream(b.data());
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 10;
      this.cellColors.saveToStream(b.data());
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 9;
      this.edgeColors.saveToStream(b.data());
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 2;
      this.colWidths.saveToStream(b.data());
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 13;
      this.colAutoSize.saveToStream(b.data());
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 12;
      this.colVisible.saveToStream(b.data());
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 21;
      this.hideDataCol.saveToStream(b.data());
      BIFF.writeBIFF(21, b, stream);
      b.reset();
      b.ident = 3;
      this.rowHeights.saveToStream(b.data());
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 15;
      this.rowAutoSize.saveToStream(b.data());
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 14;
      this.rowVisible.saveToStream(b.data());
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 22;
      this.hideDataRow.saveToStream(b.data());
      BIFF.writeBIFF(22, b, stream);
      b.reset();
      b.ident = 4;
      valueList.saveToStream(b.data());
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 11;
      this.propList.saveToStreamEx(b.data(), 0, 12);
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 19;
      this.propList.saveToStreamEx(b.data(), 12, 8);
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 1;
      propIndexs.saveToStream(b.data());
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 5;
      this.rects.saveToStream(b.data());
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 16;
      b.data().writeInt(this.scrollTopCol);
      b.data().writeInt(this.scrollTopRow);
      b.data().writeInt(this.scrollBottomCol);
      b.data().writeInt(this.scrollBottomRow);
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 17;
      b.data().writeInt(this.curCol);
      b.data().writeInt(this.curRow);
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 20;
      this.labels.saveToStream(b.data());
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 23;
      b.data().writeInt(this.colCountPerPage);
      b.data().writeInt(this.rowCountPerPage);
      b.data().writeBool(this.rowPagingFirst);
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 24;
      cellCSSList.saveToStream(b.data());
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 26;
      cellScriptList.saveToStream(b.data());
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 27;
      this.extDatas.saveToStream(b.data());
      BIFF.writeBIFF(2, b, stream);
      b.reset();
      b.ident = 127;
      BIFF.writeBIFF(2, b, stream);
   }

   public void loadFromStream(InputStream inStream) throws StreamException {
      MemStream stream = new MemStream();

      try {
         stream.loadFromStream(inStream);
      } catch (IOException var4) {
         IOException e = var4;
         throw new StreamException(e);
      }

      stream.setPosition(0L);
      this.loadFromStream((Stream)stream);
   }

   public void saveToStream(OutputStream outStream) throws StreamException {
      MemStream stream = new MemStream();
      this.saveToStream((Stream)stream);

      try {
         stream.saveToStream(outStream);
      } catch (IOException var4) {
         IOException e = var4;
         throw new StreamException(e);
      }
   }

   public void loadFromFile(String fileName) throws StreamException {
      MemStream stream = new MemStream();

      try {
         stream.loadFromFile(fileName);
      } catch (IOException var4) {
         IOException e = var4;
         throw new StreamException(e);
      }

      stream.setPosition(0L);
      this.loadFromStream((Stream)stream);
   }

   public void saveToFile(String fileName) throws StreamException {
      MemStream stream = new MemStream();
      this.saveToStream((Stream)stream);

      try {
         stream.saveToFile(fileName);
      } catch (IOException var4) {
         IOException e = var4;
         throw new StreamException(e);
      }
   }

   public GridData makeCopy() throws StreamException {
      Stream s = new MemStream();
      this.saveToStream((Stream)s);
      s.setPosition(0L);
      GridData gd = new GridData();
      gd.loadFromStream((Stream)s);
      gd.setOptions(this.getOptions());
      return gd;
   }

   public Object clone() {
      try {
         return this.makeCopy();
      } catch (StreamException var1) {
         return null;
      }
   }

   public boolean copyFrom(GridData srcData, int fromLeft, int fromTop, int fromRight, int fromBottom, int toLeft, int toTop) {
      CellField dest = new CellField(toLeft, toTop, toLeft + fromRight - fromLeft, toTop + fromBottom - fromTop);
      if (!this.checkCol(dest.left, false)) {
         return false;
      } else if (!this.checkCol(dest.right, false)) {
         return false;
      } else if (!this.checkRow(dest.top, false)) {
         return false;
      } else if (!this.checkRow(dest.bottom, false)) {
         return false;
      } else if (!srcData.checkCol(fromLeft, false)) {
         return false;
      } else if (!srcData.checkCol(fromRight, false)) {
         return false;
      } else if (!srcData.checkRow(fromTop, false)) {
         return false;
      } else if (!srcData.checkRow(fromBottom, false)) {
         return false;
      } else {
         this.rects.removeMergeRect(dest);
         GridCopyConvertor convertor = this == srcData ? null : new GridCopyConvertor(srcData);

         int x;
         int y;
         for(x = fromLeft; x <= fromRight; ++x) {
            for(y = fromTop; y <= fromBottom; ++y) {
               CellField field = srcData.expandCell(x, y);
               String text = srcData.getCellData(field.left, field.top);
               GridCell cell = srcData.getCellEx(field.left, field.top);
               if (field.left < fromLeft) {
                  field.left = fromLeft;
               }

               if (field.right > fromRight) {
                  field.right = fromRight;
               }

               if (field.top < fromTop) {
                  field.top = fromTop;
               }

               if (field.bottom > fromBottom) {
                  field.bottom = fromBottom;
               }

               if (field.left >= x && field.top >= y) {
                  int xx = toLeft + x - fromLeft;
                  int yy = toTop + y - fromTop;
                  if (convertor != null) {
                     convertor.convert(cell);
                  }

                  this.setCell(xx, yy, cell);
                  this.setCellData(xx, yy, text);
                  this.setObj(xx, yy, srcData.getObj(field.left, field.top));
                  this.setCellClass(xx, yy, srcData.getCellClass(field.left, field.top));
                  this.setCellScript(xx, yy, srcData.getCellScript(field.left, field.top));
                  this.mergeCells(xx, yy, toLeft + field.right - fromLeft, toTop + field.bottom - fromTop);
               }
            }
         }

         if (dest.top == 1 && dest.bottom == this.getRowCount() - 1) {
            for(x = fromLeft; x <= fromRight; ++x) {
               y = toLeft + x - fromLeft;
               this.setHideDataCol(y, srcData.getHideDataCol(x));
               this.setColAutoSize(y, srcData.getColAutoSize(x));
               this.setColWidths(y, srcData.getColWidths(x));
               this.setColVisible(y, srcData.getColVisible(x));
            }
         }

         if (dest.left == 1 && dest.right == this.getColCount() - 1) {
            for(x = fromTop; x <= fromBottom; ++x) {
               y = toTop + x - fromTop;
               this.setHideDataRow(y, srcData.getHideDataRow(x));
               this.setRowAutoSize(y, srcData.getRowAutoSize(x));
               this.setRowHeights(y, srcData.getRowHeights(x));
               this.setRowVisible(y, srcData.getRowVisible(x));
            }
         }

         return true;
      }
   }

   public int getColCountPerPage() {
      return this.colCountPerPage;
   }

   public void setColCountPerPage(int value) {
      this.colCountPerPage = value;
   }

   public int getRowCountPerPage() {
      return this.rowCountPerPage;
   }

   public void setRowCountPerPage(int value) {
      this.rowCountPerPage = value;
   }

   public boolean getRowPagingFirst() {
      return this.rowPagingFirst;
   }

   public void setRowPagingFirst(boolean value) {
      this.rowPagingFirst = value;
   }

   public int getColPageCount() {
      if (this.colCountPerPage <= 0) {
         return 1;
      } else {
         int colCount = this.getColCount() - 1;
         if (this.scrollBottomCol > 0 && this.scrollBottomCol > this.scrollTopCol) {
            colCount = this.scrollBottomCol;
         }

         if (this.scrollTopCol > 0) {
            colCount -= this.scrollTopCol - 1;
         }

         return (colCount - 1) / this.colCountPerPage + 1;
      }
   }

   public int getRowPageCount() {
      if (this.rowCountPerPage <= 0) {
         return 1;
      } else {
         int rowCount = this.getRowCount() - 1;
         if (this.scrollBottomRow > 0 && this.scrollBottomRow > this.scrollTopRow) {
            rowCount = this.scrollBottomRow;
         }

         if (this.scrollTopRow > 0) {
            rowCount -= this.scrollTopRow - 1;
         }

         return (rowCount - 1) / this.rowCountPerPage + 1;
      }
   }

   public int getPageCount() {
      return this.getColPageCount() * this.getRowPageCount();
   }

   public GridData getPagedGridData(int pageIndex) {
      int px = false;
      int py = false;
      int count;
      int px;
      int py;
      if (this.rowPagingFirst) {
         count = this.getRowPageCount();
         px = pageIndex / count;
         py = pageIndex % count;
      } else {
         count = this.getColPageCount();
         px = pageIndex % count;
         py = pageIndex / count;
      }

      return this.getPagedGridData(px, py);
   }

   public GridData getPagedGridData(int colPageIndex, int rowPageIndex) {
      if (colPageIndex >= 0 && rowPageIndex >= 0) {
         int colHead = this.scrollTopCol - 1;
         int colTail = this.scrollBottomCol + 1;
         int rowHead = this.scrollTopRow - 1;
         int rowTail = this.scrollBottomRow + 1;
         if (colHead < 0) {
            colHead = 0;
         }

         if (colTail <= colHead) {
            colTail = this.getColCount();
         }

         if (rowHead < 0) {
            rowHead = 0;
         }

         if (rowTail <= rowHead) {
            rowTail = this.getRowCount();
         }

         int colCountPerPage = this.colCountPerPage;
         if (colCountPerPage <= 0) {
            colCountPerPage = colTail - colHead - 1;
         }

         int rowCountPerPage = this.rowCountPerPage;
         if (rowCountPerPage <= 0) {
            rowCountPerPage = rowTail - rowHead - 1;
         }

         int dataColCount = Math.min(colCountPerPage, colTail - colPageIndex * colCountPerPage - colHead - 1);
         if (dataColCount <= 0 && colPageIndex != 0) {
            return null;
         } else {
            int fromDataCol = colHead + 1 + colPageIndex * colCountPerPage;
            int dataRowCount = Math.min(rowCountPerPage, rowTail - rowPageIndex * rowCountPerPage - rowHead - 1);
            if (dataRowCount <= 0 && rowPageIndex != 0) {
               return null;
            } else {
               int fromDataRow = rowHead + 1 + rowPageIndex * rowCountPerPage;
               if (1 + colHead + dataColCount == colTail && 1 + rowHead + dataRowCount == rowTail) {
                  return this;
               } else {
                  GridData result = new GridData(this.getOptions());
                  result.cellColors = (IntList)this.cellColors.clone();
                  result.edgeColors = (IntList)this.edgeColors.clone();
                  result.fontNames = (GridFontNames)this.fontNames.clone();
                  result.fontColors = (IntList)this.fontColors.clone();
                  result.fontSizes = (IntList)this.fontSizes.clone();
                  result.setDataClass(this.getDataClass());
                  result.setColCount(1 + colHead + dataColCount + this.getColCount() - colTail);
                  result.setRowCount(1 + rowHead + dataRowCount + this.getRowCount() - rowTail);
                  result.setLockTopCol(colHead + 1);
                  result.setLockTopRow(rowHead + 1);
                  result.setLockBottomCol(colHead + dataColCount);
                  result.setLockBottomRow(rowHead + dataRowCount);

                  int i;
                  for(i = 1; i <= colHead; ++i) {
                     result.setColWidths(i, this.getColWidths(i));
                     result.setColAutoSize(i, this.getColAutoSize(i));
                     result.setColVisible(i, this.getColVisible(i));
                     result.setHideDataCol(i, this.getHideDataCol(i));
                  }

                  for(i = 0; i < dataColCount; ++i) {
                     result.setColWidths(colHead + 1 + i, this.getColWidths(fromDataCol + i));
                     result.setColAutoSize(colHead + 1 + i, this.getColAutoSize(fromDataCol + i));
                     result.setColVisible(colHead + 1 + i, this.getColVisible(fromDataCol + i));
                     result.setHideDataCol(colHead + 1 + i, this.getHideDataCol(fromDataCol + i));
                  }

                  for(i = colTail; i < this.getColCount(); ++i) {
                     result.setColWidths(colHead + 1 + dataColCount + i - colTail, this.getColWidths(i));
                     result.setColAutoSize(colHead + 1 + dataColCount + i - colTail, this.getColAutoSize(i));
                     result.setColVisible(colHead + 1 + dataColCount + i - colTail, this.getColVisible(i));
                     result.setHideDataCol(colHead + 1 + dataColCount + i - colTail, this.getHideDataCol(i));
                  }

                  for(i = 1; i <= rowHead; ++i) {
                     result.setRowHeights(i, this.getRowHeights(i));
                     result.setRowAutoSize(i, this.getRowAutoSize(i));
                     result.setRowVisible(i, this.getRowVisible(i));
                     result.setHideDataRow(i, this.getHideDataRow(i));
                  }

                  for(i = 0; i < dataRowCount; ++i) {
                     result.setRowHeights(rowHead + 1 + i, this.getRowHeights(fromDataRow + i));
                     result.setRowAutoSize(rowHead + 1 + i, this.getRowAutoSize(fromDataRow + i));
                     result.setRowVisible(rowHead + 1 + i, this.getRowVisible(fromDataRow + i));
                     result.setHideDataRow(rowHead + 1 + i, this.getHideDataRow(fromDataRow + i));
                  }

                  for(i = rowTail; i < this.getRowCount(); ++i) {
                     result.setRowHeights(rowHead + 1 + dataRowCount + i - rowTail, this.getRowHeights(i));
                     result.setRowAutoSize(rowHead + 1 + dataRowCount + i - rowTail, this.getRowAutoSize(i));
                     result.setRowVisible(rowHead + 1 + dataRowCount + i - rowTail, this.getRowVisible(i));
                     result.setHideDataRow(rowHead + 1 + dataRowCount + i - rowTail, this.getHideDataRow(i));
                  }

                  if (colHead > 0 && rowHead > 0) {
                     result.copyFrom(this, 1, 1, colHead, rowHead, 1, 1);
                  }

                  if (dataColCount > 0 && rowHead > 0) {
                     result.copyFrom(this, fromDataCol, 1, fromDataCol + dataColCount - 1, rowHead, colHead + 1, 1);
                  }

                  if (colTail < this.getColCount() && rowHead > 0) {
                     result.copyFrom(this, colTail, 1, this.getColCount() - 1, rowHead, colHead + 1 + dataColCount, 1);
                  }

                  if (colHead > 0 && dataRowCount > 0) {
                     result.copyFrom(this, 1, fromDataRow, colHead, fromDataRow + dataRowCount - 1, 1, rowHead + 1);
                  }

                  if (dataColCount > 0 && dataRowCount > 0) {
                     result.copyFrom(this, fromDataCol, fromDataRow, fromDataCol + dataColCount - 1, fromDataRow + dataRowCount - 1, colHead + 1, rowHead + 1);
                  }

                  if (colTail < this.getColCount() && dataRowCount > 0) {
                     result.copyFrom(this, colTail, fromDataRow, this.getColCount() - 1, fromDataRow + dataRowCount - 1, colHead + 1 + dataColCount, rowHead + 1);
                  }

                  if (colHead > 0 && rowTail < this.getRowCount()) {
                     result.copyFrom(this, 1, rowTail, colHead, this.getRowCount() - 1, 1, rowHead + 1 + dataRowCount);
                  }

                  if (dataColCount > 0 && rowTail < this.getRowCount()) {
                     result.copyFrom(this, fromDataCol, rowTail, fromDataCol + dataColCount - 1, this.getRowCount() - 1, colHead + 1, rowHead + 1 + dataRowCount);
                  }

                  if (colTail < this.getColCount() && rowTail < this.getRowCount()) {
                     result.copyFrom(this, colTail, rowTail, this.getColCount() - 1, this.getRowCount() - 1, colHead + 1 + dataColCount, rowHead + 1 + dataRowCount);
                  }

                  return result;
               }
            }
         }
      } else {
         return null;
      }
   }

   public int getDataRowCount() {
      int top = this.scrollTopRow > 0 ? this.scrollTopRow : 1;
      int bottom = this.scrollBottomRow > 0 ? this.scrollBottomRow : this.getRowCount() - 1;
      int count = 0;

      for(int i = top; i <= bottom; ++i) {
         if (this.getRowVisible(i)) {
            ++count;
         }
      }

      return count;
   }

   public int getDataColCount() {
      int left = this.scrollTopCol > 0 ? this.scrollTopCol : 1;
      int right = this.scrollBottomCol > 0 ? this.scrollBottomCol : this.getColCount() - 1;
      int count = 0;

      for(int i = left; i <= right; ++i) {
         if (this.getColVisible(i)) {
            ++count;
         }
      }

      return count;
   }

   public ExtendedDatas getExtDatas() {
      return this.extDatas;
   }

   public int getHorzOffset() {
      return this.horzOffset;
   }

   public void setHorzOffset(int offset) {
      this.horzOffset = offset;
   }

   public int getVertOffset() {
      return this.vertOffset;
   }

   public void setVertOffset(int offset) {
      this.vertOffset = offset;
   }

   public boolean getHorzCenter() {
      return this.horzCenter;
   }

   public void setHorzCenter(boolean value) {
      this.horzCenter = value;
   }

   public boolean getVertCenter() {
      return this.vertCenter;
   }

   public void setVertCenter(boolean value) {
      this.vertCenter = value;
   }

   public static GridData bytesToGrid(byte[] data) {
      if (data == null) {
         return null;
      } else {
         GridData gd = new GridData();

         try {
            Stream s = new MemStream();
            s.writeBuffer(data, 0, data.length);
            s.setPosition(0L);
            gd.loadFromStream((Stream)s);
         } catch (Exception var3) {
            Exception ex = var3;
            ex.printStackTrace();
            gd = null;
         }

         return gd;
      }
   }

   public static GridData bytesToGrid(byte[] data, String encodeCharset) {
      if (data == null) {
         return null;
      } else {
         GridData gd = new GridData();

         try {
            Stream s = new MemStream();
            s.setUseEncode(true);
            s.setCharset(encodeCharset);
            s.writeBuffer(data, 0, data.length);
            s.setPosition(0L);
            gd.loadFromStream((Stream)s);
         } catch (Exception var4) {
            Exception ex = var4;
            ex.printStackTrace();
            gd = null;
         }

         return gd;
      }
   }

   public static byte[] gridToBytes(GridData grid) {
      if (grid == null) {
         return null;
      } else {
         Stream store = new MemStream();

         try {
            grid.saveToStream((Stream)store);
            return store.getBytes();
         } catch (StreamException var3) {
            StreamException ex = var3;
            ex.printStackTrace();
            return null;
         }
      }
   }

   public static byte[] gridToBytes(GridData grid, String encodeCharset) {
      if (grid == null) {
         return null;
      } else {
         Stream store = new MemStream();

         try {
            store.setUseEncode(true);
            store.setCharset(encodeCharset);
            grid.saveToStream((Stream)store);
            return store.getBytes();
         } catch (StreamException var4) {
            StreamException ex = var4;
            ex.printStackTrace();
            return null;
         }
      }
   }

   public static GridData base64ToGrid(String base64Str) {
      byte[] data = Base64.base64ToByteArray(base64Str);
      return bytesToGrid(data);
   }

   public static GridData base64ToGrid(String base64Str, String encodeCharset) {
      byte[] data = Base64.base64ToByteArray(base64Str);
      return bytesToGrid(data, encodeCharset);
   }

   public static String gridToBase64(GridData grid) {
      byte[] data = gridToBytes(grid);
      return Base64.byteArrayToBase64(data);
   }

   public static String gridToBase64(GridData grid, String encodeCharset) {
      byte[] data = gridToBytes(grid, encodeCharset);
      return Base64.byteArrayToBase64(data);
   }

   public void convertData(CellConvertor convertor) {
      for(int col = 0; col < this.getColCount(); ++col) {
         for(int row = 0; row < this.getRowCount(); ++row) {
            CellField cf = this.expandCell(col, row);
            if (cf.left == col && cf.top == row) {
               GridCell cell = this.getCellEx(col, row);
               if (convertor.convert(this, cell)) {
                  this.setCell(col, row, cell);
               }
            }
         }
      }

   }

   public int getOptions() {
      return this.options;
   }

   public void setOptions(int value) {
      this.options = value;
      this.rects.setAutoMerge((value & 1) != 0);
   }

   private final class GridCopyConvertor {
      GridData src;
      int[] fontNameMaps;
      int[] fontSizeMaps;
      int[] fontColorMaps;
      int[] cellColorMaps;
      int[] edgeColorMaps;
      private Stack freeFontNames;
      private Stack freeFontSizes;
      private Stack freeFontColors;
      private Stack freeCellColors;
      private Stack freeEdgeColors;

      public GridCopyConvertor(GridData src) {
         this.src = src;
         this.mergeFontNames(src.fontNames);
         this.fontSizeMaps = this.mergeList(src.fontSizes, GridData.this.fontSizes);
         this.fontColorMaps = this.mergeList(src.fontColors, GridData.this.fontColors);
         this.cellColorMaps = this.mergeList(src.cellColors, GridData.this.cellColors);
         this.edgeColorMaps = this.mergeList(src.edgeColors, GridData.this.edgeColors);
      }

      private void mergeFontNames(GridFontNames srcFontNames) {
         boolean diff = false;
         int[] maps = new int[srcFontNames.count()];

         for(int i = 0; i < srcFontNames.count(); ++i) {
            int index = GridData.this.fontNames.indexOfFont(srcFontNames.getFont(i));
            maps[i] = index;
            if (i != index) {
               diff = true;
            }
         }

         this.fontNameMaps = diff ? maps : null;
      }

      private int[] mergeList(IntList srcList, IntList destList) {
         boolean diff = false;
         int[] maps = new int[srcList.count()];

         for(int i = 0; i < srcList.count(); ++i) {
            int index = destList.indexOf(srcList.get(i));
            maps[i] = index;
            if (i != index) {
               diff = true;
            }
         }

         return diff ? maps : null;
      }

      public void convert(GridCell cell) {
         byte[] data = cell.getPropData();
         int index;
         int newIndex;
         if (this.fontNameMaps != null) {
            index = data[4] & 63;
            newIndex = this.mapFontName(index);
            data[4] = (byte)(data[4] & 192 | newIndex & 63);
         }

         if (this.fontSizeMaps != null) {
            index = data[5] & 63;
            newIndex = this.mapFontSize(index);
            data[5] = (byte)(data[5] & 192 | newIndex & 63);
         }

         if (this.fontColorMaps != null) {
            index = data[6] >> 0 & 63;
            newIndex = this.mapFontColor(index);
            data[6] = (byte)(data[6] & 192 | newIndex & 63);
         }

         if (this.cellColorMaps != null) {
            int indexx = data[0];
            newIndex = this.mapCellColor(indexx);
            data[0] = (byte)newIndex;
         }

         if (this.edgeColorMaps != null) {
            index = data[9] >> 4 & 15;
            newIndex = this.mapEdgeColor(index);
            data[9] = (byte)(data[9] & 15 | newIndex << 4 & 240);
            index = data[10] >> 4 & 15;
            newIndex = this.mapEdgeColor(index);
            data[10] = (byte)(data[10] & 15 | newIndex << 4 & 240);
         }

      }

      private int mapCellColor(int index) {
         int newIndex;
         if (this.cellColorMaps[index] == -1) {
            if (this.freeCellColors().isEmpty()) {
               newIndex = GridData.this.cellColors.count() >= 64 ? 0 : GridData.this.cellColors.add(this.src.cellColors.get(index));
            } else {
               newIndex = (Integer)this.freeCellColors().pop();
               GridData.this.cellColors.set(newIndex, this.src.cellColors.get(index));
            }

            this.cellColorMaps[index] = newIndex;
         } else {
            newIndex = this.cellColorMaps[index];
         }

         return newIndex;
      }

      private int mapFontColor(int index) {
         int newIndex;
         if (this.fontColorMaps[index] == -1) {
            if (this.freeFontColors().isEmpty()) {
               newIndex = GridData.this.fontColors.count() >= 64 ? 0 : GridData.this.fontColors.add(this.src.fontColors.get(index));
            } else {
               newIndex = (Integer)this.freeFontColors().pop();
               GridData.this.fontColors.set(newIndex, this.src.fontColors.get(index));
            }

            this.fontColorMaps[index] = newIndex;
         } else {
            newIndex = this.fontColorMaps[index];
         }

         return newIndex;
      }

      private int mapFontSize(int index) {
         int newIndex;
         if (this.fontSizeMaps[index] == -1) {
            if (this.freeFontSizes().isEmpty()) {
               newIndex = GridData.this.fontSizes.count() >= 64 ? 0 : GridData.this.fontSizes.add(this.src.fontSizes.get(index));
            } else {
               newIndex = (Integer)this.freeFontSizes().pop();
               GridData.this.fontSizes.set(newIndex, this.src.fontSizes.get(index));
            }

            this.fontSizeMaps[index] = newIndex;
         } else {
            newIndex = this.fontSizeMaps[index];
         }

         return newIndex;
      }

      private int mapFontName(int index) {
         int newIndex;
         if (this.fontNameMaps[index] == -1) {
            if (this.freeFontNames().isEmpty()) {
               newIndex = GridData.this.fontNames.count() >= 64 ? 0 : GridData.this.fontNames.addFont(this.src.fontNames.getFont(index));
            } else {
               newIndex = (Integer)this.freeFontNames().pop();
               GridData.this.fontNames.setFont(newIndex, this.src.fontNames.getFont(index));
            }

            this.fontNameMaps[index] = newIndex;
         } else {
            newIndex = this.fontNameMaps[index];
         }

         return newIndex;
      }

      private int mapEdgeColor(int index) {
         int newIndex;
         if (this.edgeColorMaps[index] == -1) {
            if (this.freeEdgeColors().isEmpty()) {
               newIndex = GridData.this.edgeColors.count() >= 64 ? 0 : GridData.this.edgeColors.add(this.src.edgeColors.get(index));
            } else {
               newIndex = (Integer)this.freeEdgeColors().pop();
               GridData.this.edgeColors.set(newIndex, this.src.edgeColors.get(index));
            }

            this.edgeColorMaps[index] = newIndex;
         } else {
            newIndex = this.edgeColorMaps[index];
         }

         return newIndex;
      }

      private Stack freeFontNames() {
         if (this.freeFontNames == null) {
            boolean[] flags = new boolean[GridData.this.fontNames.count()];

            int i;
            for(i = 0; i < GridData.this.propList.count(); ++i) {
               byte[] data = GridData.this.propList.get(i);
               int index = data[4] & 63;
               flags[index] = true;
            }

            this.freeFontNames = new Stack();

            for(i = 0; i < flags.length; ++i) {
               if (!flags[i]) {
                  this.freeFontNames.push(new Integer(i));
               }
            }
         }

         return this.freeFontNames;
      }

      private Stack freeFontSizes() {
         if (this.freeFontSizes == null) {
            boolean[] flags = new boolean[GridData.this.fontSizes.count()];

            int i;
            for(i = 0; i < GridData.this.propList.count(); ++i) {
               byte[] data = GridData.this.propList.get(i);
               int index = data[5] & 63;
               flags[index] = true;
            }

            this.freeFontSizes = new Stack();

            for(i = 0; i < flags.length; ++i) {
               if (!flags[i]) {
                  this.freeFontSizes.push(new Integer(i));
               }
            }
         }

         return this.freeFontSizes;
      }

      private Stack freeFontColors() {
         if (this.freeFontColors == null) {
            boolean[] flags = new boolean[GridData.this.fontColors.count()];

            int i;
            for(i = 0; i < GridData.this.propList.count(); ++i) {
               byte[] data = GridData.this.propList.get(i);
               int index = data[6] >> 0 & 63;
               flags[index] = true;
            }

            this.freeFontColors = new Stack();

            for(i = 0; i < flags.length; ++i) {
               if (!flags[i]) {
                  this.freeFontColors.push(new Integer(i));
               }
            }
         }

         return this.freeFontColors;
      }

      private Stack freeCellColors() {
         if (this.freeCellColors == null) {
            boolean[] flags = new boolean[GridData.this.cellColors.count()];

            int i;
            for(i = 0; i < GridData.this.propList.count(); ++i) {
               byte[] data = GridData.this.propList.get(i);
               flags[data[0]] = true;
            }

            this.freeCellColors = new Stack();

            for(i = 0; i < flags.length; ++i) {
               if (!flags[i]) {
                  this.freeCellColors.push(new Integer(i));
               }
            }
         }

         return this.freeCellColors;
      }

      private Stack freeEdgeColors() {
         if (this.freeEdgeColors == null) {
            boolean[] flags = new boolean[GridData.this.edgeColors.count()];

            int i;
            for(i = 0; i < GridData.this.propList.count(); ++i) {
               byte[] data = GridData.this.propList.get(i);
               int index = data[9] >> 4 & 15;
               flags[index] = true;
               index = data[10] >> 4 & 15;
               flags[index] = true;
            }

            this.freeEdgeColors = new Stack();

            for(i = 0; i < flags.length; ++i) {
               if (!flags[i]) {
                  this.freeEdgeColors.push(new Integer(i));
               }
            }
         }

         return this.freeEdgeColors;
      }
   }
}
